unit ID3Run;
interface

function ParseOptions: Boolean;
procedure ExecuteCommands;
procedure ShowHelp;

implementation
uses SysUtils, ID3;

{ ID3 Genre management }
function ID3NameToGenre(AName: String): ID3Genre;
var RetVal: ID3Genre;
begin
  Result := ID3UnknowGenre;
  for RetVal := Low(ID3Genre) to High(ID3Genre) do
    if UpperCase(AName) = UpperCase(StrPas(ID3GenreName[RetVal])) then
      begin
        Result := RetVal;
        Break;
      end;
end;

{ Parameter management }
const NextParamCount: Integer = 1;

function HasMoreParams: Boolean;
begin
  Result := NextParamCount <= ParamCount;
end;

procedure FirstParam;
begin
  NextParamCount := 1;
end;

function NextParam: String;
begin
  Result := ParamStr(NextParamCount);
  Inc(NextParamCount);
  if (Result <> '') and (Result[1] = '"') then
    begin
      repeat
        Result := Result + ParamStr(NextParamCount);
        Inc(NextParamCount);
      until not HasMoreParams and (Copy(Result, Length(Result), 1) = '"');
      Result := Copy(Result, 2, Length(Result) - 2);
    end;
end;

{ Program behaviour }
const
  FileName: String = '*.mp3';
  WriteTags: Boolean = false;
  DeleteTags: Boolean = false;
  Verbose: Boolean = false;
  ListIt: Boolean = true;
  NewTitle: String = '';
  NewArtist: String = '';
  NewAlbum: String = '';
  NewYear: String = '';
  NewComment: String = '';
  NewGenre: ID3Genre = ID3UnknowGenre;
  NewFileName: String = '';
  FNPattern: String = '';

function ParseOptions: Boolean;
var
  Param: String;
begin
  Result := true;
  FirstParam;
  while Result and HasMoreParams do
    begin
      Param := NextParam;
      if Param <> '' then
        if Param[1] in ['-', '/'] then
          case UpCase(Param[2]) of
            'H': Result := false;
            'D': DeleteTags := true;
            'V': Verbose := true;
            'N': ListIt := false;
            'I': FNPattern := NextParam;
            'B': NewArtist := NextParam;
            'A': NewAlbum := NextParam;
            'T': NewTitle := NextParam;
            'Y': NewYear := NextParam;
            'C': NewComment := NextParam;
            'F': NewFileName := NextParam;
            'G': NewGenre := ID3NameToGenre(NextParam);
          end
        else
          FileName := Param;
    end;
  WriteTags := (NewTitle <> '') or (NewArtist <> '') or
        (NewAlbum <> '') or (NewYear <> '') or (NewComment <> '') or
        (NewGenre <> ID3UnknowGenre);
end;

function Interpret(FN: String; var Tag: ID3Struct): String;
var
  Parts: Array [1..10] of String;
  I, J, CurrentPart, SrcI: Integer;
  Delimiter: Char;
  InPart: Boolean;

function Cooked(S: String): String;
var
  I: Integer;
  Replacement: String;
const
  UniqueID: Integer = 1;
begin
  Result := '';
  I := 1;
  while I <= Length(S) do
    begin
      Replacement := '';
      if (S[I] = '@') and (I <  Length(S)) then
        begin
          case UpCase(S[I + 1]) of
            '1'..'9':
              if Ord(S[I + 1]) - Ord('0') <= CurrentPart then
                Replacement := Parts[Ord(S[I + 1]) - Ord('0')];
            'B': Replacement := Trim(Tag.Artist);
            'A': Replacement := Trim(Tag.Album);
            'T': Replacement := Trim(Tag.Title);
            'Y': Replacement := Trim(Tag.Year);
            'C': Replacement := Trim(Tag.Comment);
            '@': Replacement := '@';
            '#': begin
                Replacement := Format('%2.2d', [UniqueID]);
                Inc(UniqueID);
              end;
          end;
          Inc(I);
        end
      else
        Replacement := S[I];
      Result := Result + Replacement;
      Inc(I);
    end;
end;

begin
  CurrentPart := 0;
  InPart := false;
  SrcI := 1;
  for I := 1 to Length(FNPattern) do
    case FNPattern[I] of
      '@': begin
          InPart := not InPart;
          if InPart then
            begin
              Inc(CurrentPart);
              Parts[CurrentPart] := '';
            end;
        end;
      '*': begin
          J := I + 1;
          while (J <= Length(FNPattern)) and (FNPattern[J] = '@') do
            Inc(J);
          if J <= Length(FNPattern) then Delimiter := FNPattern[J]
                                    else Delimiter := '*'; { matches rest! }
          while (SrcI <= Length(FN)) and (FN[SrcI] <> Delimiter) do
            begin
              if InPart then
                Parts[CurrentPart] := Parts[CurrentPart] + FN[SrcI];
              Inc(SrcI);
            end;
        end;
      '?': begin
          if InPart then
            Parts[CurrentPart] := Parts[CurrentPart] + FN[SrcI];
          Inc(SrcI);
        end;
      else begin
          if FNPattern[I] = FN[SrcI] then Inc(SrcI); {else error!}
        end;
    end;
  if NewTitle    <> '' then StrPLCopy(Tag.Title  , Cooked(NewTitle),   30);
  if NewArtist   <> '' then StrPLCopy(Tag.Artist , Cooked(NewArtist),  30);
  if NewAlbum    <> '' then StrPLCopy(Tag.Album  , Cooked(NewAlbum),   30);
  if NewYear     <> '' then StrPLCopy(Tag.Year   , Cooked(NewYear),     4);
  if NewComment  <> '' then StrPLCopy(Tag.Comment, Cooked(NewComment), 30);
  if NewGenre <> ID3UnknowGenre then Tag.Genre := NewGenre;
  if NewFileName <> '' then Result := Cooked(NewFileName) else Result := FN;
end;

procedure ProcessFile(FN: String);
var
  F: File of Byte;
  Tag: ID3Struct;
  HasTag: Boolean;
  GenreName: String;
  NewFN: String;
const
  Tab = #9;
begin
  AssignFile(F, FN);
  try
    Reset(F);
    try
      Seek(F, FileSize(F) - ID3OffsetFromEnd);
      BlockRead(F, Tag, SizeOf(Tag));
      HasTag := Tag.Magic = ID3Magic;
      if not HasTag then
        begin
          FillChar(Tag, Sizeof(Tag), ' ');
          Tag.Magic := ID3Magic;
          Tag.Genre := ID3UnknowGenre;
        end;

      NewFN := ExtractFilePath(FN) +
               Interpret(ChangeFileExt(ExtractFileName(FN), ''), Tag) +
               ExtractFileExt(FN);

      if ListIt then
        with Tag do
          begin
            if Genre <= High(ID3Genre) then
              GenreName := ID3GenreName[Genre]
            else
              GenreName := 'Unknown: ' + IntToStr(Ord(Genre));
            Writeln(ExtractFileName(FN), Tab,
                    Trim(Artist), Tab, Trim(Title), Tab, Trim(Album),
                    Tab, GenreName);
          end;

      if Verbose then Write('File ', FN, ':');
      if DeleteTags then
        begin
          if HasTag then
            begin
              Seek(F, FileSize(F) - ID3OffsetFromEnd);
              Truncate(F);
              if Verbose then Write(' Tags deleted');
            end;
        end
      else if WriteTags then
        begin
          if HasTag then Seek(F, FileSize(F) - ID3OffsetFromEnd)
                    else Seek(F, FileSize(F));
          BlockWrite(F, Tag, SizeOf(Tag));
          if Verbose then
            if HasTag then Write(' Tags modified') else Write(' Tags added');
        end;
    finally
      CloseFile(F);
    end;
    if (NewFN <> FN) then
      begin
        RenameFile(FN, NewFN);
        if Verbose then Write(' renamed to ', NewFN);
      end;
    if Verbose then Writeln;
  except on e: EInOutError do
  end;
end;

procedure ExecuteCommands;
var
  GoOn: Integer;
  SR: TSearchRec;
  Path: String;
begin
  Path := ExtractFilePath(FileName);
  GoOn := FindFirst(FileName, faAnyFile, SR);
  while GoOn = 0 do
    begin
      if (SR.Attr and (faDirectory or faVolumeID)) = 0 then
        ProcessFile(Path + SR.Name);
      GoOn := FindNext(SR);
    end;
  FindClose(SR);
end;

procedure ShowHelp;
var GI: ID3Genre;
begin
  Writeln(ExtractFileName(ParamStr(0)) + ' [-H] <File(s)>');
  Writeln;
  Writeln('Options');
  Writeln('  -H           Show this help');
  Writeln('  -V           Verbose operation');
  Writeln('  -N           Don''t show each file');
  Writeln('  -I <pattern> Interprets file name as <pattern>, where <pattern> can');
  Writeln('               have sections surrounded by "@", which can be used for');
  Writeln('               replacements in any ID3 tag');
  Writeln('  -D           Delete tags');
  Writeln('  -B <artist>  Artist tag (can have @<n> replacements)');
  Writeln('  -A <album>   Album tag (can have @<n> replacements)');
  Writeln('  -T <title>   Song title tag (can have @<n> replacements)');
  Writeln('  -Y <year>    Year tag (can have @<n> replacements)');
  Writeln('  -C <comment> Comment tag (can have @<n> replacements)');
  Writeln('  -G <genre>   Genre tag (Do -V -H to see all valid genres)');
  Writeln('  -F <file>    Filename to rename to (can have any replacements)');
  Writeln;
  Writeln('Patterns');
  Writeln('  Every tag field, as well as the rename filename, can have the');
  Writeln('following replacement specifications:');
  Writeln('  @1 to @9     Is replaced by the original filename matched section');
  Writeln('  @#           Provides an autoincremented counter (2 digits)');
  Writeln('  @B           Is replaced by the Artist tag contents');
  Writeln('  @A           Is replaced by the Album tag contents');
  Writeln('  @T           Is replaced by the Title tag contents');
  Writeln('  @Y           Is replaced by the Year tag contents');
  Writeln('  @C           Is replaced by the Comment tag contents');
  Writeln('  @@           Is replaced by "@"');
  Writeln('');
  if Verbose then
    begin
      Writeln('Valid Genres');
      for GI := Low(ID3Genre) to High(ID3Genre) do
        Write(StrPas(ID3GenreName[GI]), #9);
      Writeln;
    end;
end;

end.
